#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>

#include "nist.h"

#define NRUN 1000000

void print_FFT(fft_t *X) {
  
  int j;	
	
  for (j=0; j<128; j++) {
/*     if (X[j] > 128 || X[j] < -128) */
/*       printf ("\nUnreduced value!  %3i: %i\n", j, X[j]); */
    printf ("%04x ", (X[j]+257)%257);
    if (j%8 == 7)
      printf ("\n");
  }
}

void naive_dft(fft_t *x, unsigned char *a) {

  /*
   * FFT_128 using w=139 as 128th root of unity.
   *  Decimation in frequency (DIF) NTT.
   *  Output data is in revbin_permuted order.
   *  In place.
   */

  const int alpha = 139;
  int alpha_i = 1; /* this contains alpha^i; */
  int beta = 98;
  int beta_i = 1;
  int i, j;

  for(i=0; i<128; i++) {
    int alpha_ij = 1; /* this contains alpha^(i*j); */

    x[i] = beta_i;
    beta_i = (beta_i * beta) % 257;
  
    for(j=0; j<64; j++) {
      x[i] = (x[i] + alpha_ij * a[j]) % 257;
      alpha_ij = (alpha_ij * alpha_i) % 257;
    }
    
    alpha_i = (alpha_i * alpha) % 257;
  }
}

void print_state(hashState *s) {
  unsigned int i;

  for (i=0; i<s->n_feistels; i++) {
    printf ("%08x %08x %08x %08x\n", s->A[i], s->B[i], s->C[i], s->D[i]);
  }
}

int main(int argc, char* argv[]) {

  unsigned char A[128] __attribute__ ((aligned)) = {0};

  hashState s;
#ifdef rdtsc
  u64 tsc;
#endif
  struct timeval tv1;
  struct timeval tv2;
  float dt;

  int i, j;
  u32 c;

  /*
   * Check FFT against naive implementation
   */
  if (argc > 1 && argv[1][0] == 'c') {
    fft_t X1[128] __attribute__ ((aligned)) = {0};
    fft_t X2[128] __attribute__ ((aligned)) = {0};

    printf ("Check FFT Implementation\n");

    srand(getpid());
    for(c = 0; c < 0x10000; c++) {
      for (j=0; j<64; j++)
        A[j] = rand() & 0xff;

      naive_dft(X2, A);
      fft128_natural(X1, A);

      for (j=0; j<128; j++)
        if ((X2[j] - X1[j])%257) {
          printf ("\nCheck Failed!\n\nFFT128:\n");
          print_FFT(X1);
          printf ("\nNaive FFT:\n");
          print_FFT(X2);
          exit(-1);
        }
    }
    printf ("Success on 2^16 random FFT\n");
    return 0;
  }

  printf("THIS IS VERSION: %s\n", VERSION());


  /* Check implemenation using a deterministic random intialisation */
  if (argc > 1)
    srand48(atoi(argv[1]));

  for (i=0; i<128; i++) {
    A[i] = mrand48();
  }
  
  /** now benches speed ****/

  printf("SIMD-256:\n===============\n");

  Init(&s, 256);

  Update(&s, A, sizeof(A));
  Final(&s, A);
  print_state(&s);

  gettimeofday(&tv1, NULL);
#ifdef rdtsc
  tsc = rdtsc();
#endif

  for(i=0; i<NRUN; i++)
    SIMD_Compress(&s, A, 0);

#ifdef rdtsc
  tsc = rdtsc() - tsc;
#endif

  gettimeofday(&tv2, NULL);
 
  dt = (tv2.tv_sec + tv2.tv_usec/1000000.0) -
    (tv1.tv_sec + tv1.tv_usec/1000000.0);
  
  printf ("Time : %g s           (gettimeofday)\n", dt);
#ifdef rdtsc
  printf ("Speed: %g cycles/byte (rdtsc)\n", (float)tsc/NRUN/64);
#endif
  printf ("Speed: %g Mbyte/s     (gettimeofday)\n", ((float)NRUN/1000000)*64/dt);


  if (Init(&s, 512) == BAD_HASHBITLEN)
    return 0;

  Update(&s, A, sizeof(A));
  Final(&s, A);
  print_state(&s);

  printf("SIMD-512:\n===============\n");

  for (i=0; i<8; i++) {
    s.A[i] = mrand48();
    s.B[i] = mrand48();
    s.C[i] = mrand48();
    s.D[i] = mrand48();
  }

  gettimeofday(&tv1, NULL);
#ifdef rdtsc
  tsc = rdtsc();
#endif

  for(i=0; i<NRUN; i++)
    SIMD_Compress(&s, A, 0);

#ifdef rdtsc
  tsc = rdtsc() - tsc;
#endif

  gettimeofday(&tv2, NULL);
 
  dt = (tv2.tv_sec + tv2.tv_usec/1000000.0) -
    (tv1.tv_sec + tv1.tv_usec/1000000.0);
  
  printf ("Time : %g s           (gettimeofday)\n", dt);
#ifdef rdtsc
  printf ("Speed: %g cycles/byte (rdtsc)\n", (float)tsc/NRUN/128);
#endif
  printf ("Speed: %g Mbyte/s     (gettimeofday)\n", ((float)NRUN/1000000)*128/dt);

  return 0;
}
