#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "nist.h"

#define NRUN 50000

void fft128_natural(fft_t *a, unsigned char *x);
void SIMD_Compress(hashState * state, const unsigned char *M);
char* VERSION(void);

void print_FFT(fft_t *X) {
  
  int j;	
	
  for (j=0; j<128; j++) {
    if (X[j] > 128 || X[j] < -128)
      printf ("\nUnreduced value!  %3i: %i\n", j, X[j]);
    printf ("%04x ", (X[j]+257)%257);
    if (j%8 == 7)
      printf ("\n");
  }
}

void naive_dft(fft_t *x, unsigned char *a) {

  /*
   * FFT_128 using w=139 as 128th root of unity.
   *  Decimation in frequency (DIF) NTT.
   *  Output data is in revbin_permuted order.
   *  In place.
   */

  const int alpha = 139;
  int alpha_i = 1; /* this contains alpha^i; */
  int beta = 96;
  int beta_i = beta;
  int i, j;

  for(i=0; i<128; i++) {
    int alpha_ij = 1; /* this contains alpha^(i*j); */

    x[i] = beta_i;
    beta_i = (beta_i * beta) % 257;
  
    for(j=0; j<64; j++) {
      x[i] = (x[i] + alpha_ij * a[j]) % 257;
      alpha_ij = (alpha_ij * alpha_i) % 257;
    }
    
    alpha_i = (alpha_i * alpha) % 257;
  }
}

void print_state(hashState *s) {
  int i;

  for (i=0; i<s->n_feistels; i++) {
    printf ("%08x %08x %08x %08x\n", s->A[i], s->B[i], s->C[i], s->D[i]);
  }
}

int main(int argc, char* argv[]) {
  // unsigned char A[128] = {0};
  unsigned char __A[128+16];
  unsigned char *_A = __A;
  unsigned char *A;

  hashState s;
#ifdef rdtsc
  u64 tsc;
#endif
  clock_t t1, t2;
  float dt;
  int i, j;
  u32 c;

  /*
   * Align the buffer to a 128 bit boundary.
   */
  _A += (_A - (unsigned char*)NULL)&15;
  A   = (unsigned char*)_A;

  printf("sizeof(char*) = %i\n", sizeof(char*));

  // Check FFT against naive implementation
  if (argc > 1 && argv[1][0] == 'c') {
    fft_t X1[128] = {0};
    fft_t X2[128] = {0};

    printf ("Check FFT Implementation\n");

    srand((int)time(NULL));
    for(c = 0; c < 0x10000; c++) {
      for (j=0; j<64; j++)
        A[j] = rand() & 0xff;

      naive_dft(X2, A);
      fft128_natural(X1, A);

      for (j=0; j<128; j++)
        if ((X2[j] - X1[j])%257) {
          printf ("\nCheck Failed!\n\nFFT128:\n");
          print_FFT(X1);
          printf ("\nNaive FFT:\n");
          print_FFT(X2);
          exit(-1);
        }
    }
    printf ("Success on 2^16 random FFT\n");
    return 0;
  }

  printf("THIS IS VERSION: %s\n", VERSION());


  /* Check implemenation using a deterministic random intialisation */
  if (argc > 1)
    srand(atoi(argv[1]));

  for (i=0; i<128; i++) {
    A[i] = rand();
  }
  
  /** now benches speed ****/

  printf("SIMD-256:\n===============\n");

  Init(&s, 256);

  for (i=0; i<4; i++) {
    s.A[i] = rand();
    s.B[i] = rand();
    s.C[i] = rand();
    s.D[i] = rand();
  }

#ifdef rdtsc
  tsc = rdtsc();
#endif
  t1 = clock();

  for(i=0; i<NRUN; i++)
    SIMD_Compress(&s, A);

#ifdef rdtsc
  tsc = rdtsc() - tsc;
#endif
  t2 = clock();
  dt = (float) (t2 - t1) / CLOCKS_PER_SEC;

  printf ("Time : %g s           (clock)\n", dt);
  printf ("Speed: %g cycles/byte (rdtsc)\n", (float)tsc/NRUN/64);
  printf ("Speed: %g Mbyte/s     (clock)\n", ((float)NRUN/1000000)*64/dt);


  if (Init(&s, 512) == BAD_HASHBITLEN) {
	getchar();
	return 0;	
  }

  printf("SIMD-512:\n===============\n");

  for (i=0; i<8; i++) {
    s.A[i] = rand();
    s.B[i] = rand();
    s.C[i] = rand();
    s.D[i] = rand();
  }

#ifdef rdtsc
  tsc = rdtsc();
#endif
  t1 = clock();

  for(i=0; i<NRUN; i++)
    SIMD_Compress(&s, A);

#ifdef rdtsc
  tsc = rdtsc() - tsc;
#endif
  t2 = clock();
  dt = (float) (t2 - t1) / CLOCKS_PER_SEC;
 
  printf ("Time : %g s           (clock)\n", dt);
  printf ("Speed: %g cycles/byte (rdtsc)\n", (float)tsc/NRUN/128);
  printf ("Speed: %g Mbyte/s     (clock)\n", ((float)NRUN/1000000)*128/dt);

  getchar();
  return 0;
}
